/*
* Copyright (c) 2007, Carnegie Mellon University
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of University of Southern California, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY CARNEGIE MELLON UNIVERSITY
* ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
* IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
* Copyright (c) 2015, University of Southern California
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of University of Southern California, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY UNIVERSITY OF SOUTHERN CALIFORNIA 
* ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
* IN NO EVENT SHALL UNIVERSITY OF SOUTHERN CALIFORNIA BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
  Code author: Jernej Barbic, Hongyi Xu, Yijing Li

  Multithreaded computation of the distance field.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include "distanceFieldMT.h"

DistanceFieldMT::DistanceFieldMT(int numThreads_) : DistanceField(), numThreads(numThreads_) 
{
}

DistanceFieldMT::DistanceFieldMT(int resolutionX_, int resolutionY_, int resolutionZ_, int numThreads_): DistanceField(resolutionX_, resolutionY_, resolutionZ_), numThreads(numThreads_)
{
}

void DistanceFieldMT::Setup()
{
  // split the workload
  startSlice = (int*) malloc (sizeof(int) * numThreads);
  endSlice = (int*) malloc (sizeof(int) * numThreads);

  int totalSlices = (zMax - zMin + 1);

  int remainder = totalSlices % numThreads;
  // the first 'remainder' nodes will process one element more
  int jobSize = totalSlices / numThreads;

  for(int rank=0; rank < numThreads; rank++)
  {
    if (rank < remainder)
    {
      startSlice[rank] = zMin + rank * (jobSize+1);
      endSlice[rank] = zMin + (rank+1) * (jobSize+1);
    }
    else
    {
      startSlice[rank] = zMin + remainder * (jobSize+1) + (rank-remainder) * jobSize;
      endSlice[rank] = zMin + remainder * (jobSize+1) + ((rank-remainder)+1) * jobSize;
    }
  }

  printf("Total slices: %d \n", totalSlices);
  printf("Num threads: %d \n", numThreads);
  printf("Canonical job size: %d \n", jobSize);
  printf("Num threads with job size augmented by one element: %d \n", remainder);
}

DistanceFieldMT::~DistanceFieldMT()
{
}

int DistanceFieldMT::GetStartSlice(int rank)
{
  return startSlice[rank];
}

int DistanceFieldMT::GetEndSlice(int rank)
{
  return endSlice[rank];
}

// the routines for signed and unsigned distance field computation
#define COMPUTE_SIGNED_FIELD
  #define COMPUTE_FLOOD_FIELD
    #include "computeFieldMT.cpp"
  #undef COMPUTE_FLOOD_FIELD
    #include "computeFieldMT.cpp"
#undef COMPUTE_SIGNED_FIELD
  #include "computeFieldMT.cpp"

