/*
* Copyright (c) 2007, Carnegie Mellon University
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of University of Southern California, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY CARNEGIE MELLON UNIVERSITY
* ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
* IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
* Copyright (c) 2015, University of Southern California
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of University of Southern California, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY UNIVERSITY OF SOUTHERN CALIFORNIA 
* ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
* IN NO EVENT SHALL UNIVERSITY OF SOUTHERN CALIFORNIA BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
  Code author: Jernej Barbic
  Trilinear interpolation.
*/

#define TRILINEAR_INTERPOLATION(wx,wy,wz,v000,v100,v110,v010,v001,v101,v111,v011) \
( (wx) * (wy) * (wz) *             (v111) + \
  (wx) * (wy) * (1-(wz)) *         (v110) + \
  (wx) * (1-(wy)) * (wz) *         (v101) + \
  (wx) * (1-(wy)) * (1-(wz)) *     (v100) + \
  (1-(wx)) * (wy) * (wz) *         (v011) + \
  (1-(wx)) * (wy) * (1-(wz)) *     (v010) + \
  (1-(wx)) * (1-(wy)) * (wz) *     (v001) + \
  (1-(wx)) * (1-(wy)) * (1-(wz)) * (v000))

#define GRADIENT_COMPONENT_X(wx,wy,wz,v000,v100,v110,v010,v001,v101,v111,v011) \
  (((wy) * (wz) *             (v111) + \
    (wy) * (1-(wz)) *         (v110) + \
    (1-(wy)) * (wz) *         (v101) + \
    (1-(wy)) * (1-(wz)) *     (v100) + \
    (-1) * (wy) * (wz) *         (v011) + \
    (-1) * (wy) * (1-(wz)) *     (v010) + \
    (-1) * (1-(wy)) * (wz) *     (v001) + \
    (-1) * (1-(wy)) * (1-(wz)) * (v000) ) / gridX)

#define GRADIENT_COMPONENT_Y(wx,wy,wz,v000,v100,v110,v010,v001,v101,v111,v011) \
  (((wx) * (wz) *             (v111) + \
    (wx) * (1-(wz)) *         (v110) + \
    (wx) * (-1) * (wz) *         (v101) + \
    (wx) * (-1) * (1-(wz)) *     (v100) + \
    (1-(wx)) * (wz) *         (v011) + \
    (1-(wx)) * (1-(wz)) *     (v010) + \
    (1-(wx)) * (-1) * (wz) *     (v001) + \
    (1-(wx)) * (-1) * (1-(wz)) * (v000)) / gridY)
  
#define GRADIENT_COMPONENT_Z(wx,wy,wz,v000,v100,v110,v010,v001,v101,v111,v011) \
  (((wx) * (wy) *                (v111) + \
    (wx) * (wy) * (-1) *         (v110) + \
    (wx) * (1-(wy)) *            (v101) + \
    (wx) * (1-(wy)) * (-1) *     (v100) + \
    (1-(wx)) * (wy) *            (v011) + \
    (1-(wx)) * (wy) * (-1) *     (v010) + \
    (1-(wx)) * (1-(wy)) *        (v001) + \
    (1-(wx)) * (1-(wy)) * (-1) * (v000)) / gridZ)

